//! Module for converting XML custom token string to rust types.
//!
//! Unlike, [`crate::converter`], this only has string "parser".

use std::any;

pub(crate) use parse_custom_bool::ParseCustomBool;
pub(crate) use parse_prefixed_u32::ParsePrefixedU32;
pub use parse_prefixed_u32::ParsePrefixedU32Error;
use thiserror::Error;

#[derive(Debug, Error, PartialEq)]
pub enum ParseEnumError {
    #[error("{found} is not a valid variant of {type}")]
    InvalidVariant { found: String, r#type: &'static str },
}

impl ParseEnumError {
    pub fn invalid_variant<T>(found: String) -> Self {
        Self::InvalidVariant {
            found,
            r#type: any::type_name::<T>(),
        }
    }
}

mod parse_prefixed_u32 {
    use std::num;

    use thiserror::Error;

    #[derive(Debug, Error, PartialEq)]
    pub enum ParsePrefixedU32Error {
        #[error("invalid prefix: found '{found}'")]
        InvalidPrefix { found: char },
        #[error("parse int error")]
        ParseIntError(#[from] num::ParseIntError),
    }

    pub(crate) trait ParsePrefixedU32 {
        fn parse_prefixed_u32(&self) -> Result<u32, ParsePrefixedU32Error>;
    }

    impl ParsePrefixedU32 for &str {
        fn parse_prefixed_u32(&self) -> Result<u32, ParsePrefixedU32Error> {
            let radix = match self.chars().nth(1) {
                Some('x') => 16,
                Some(c) if c.is_numeric() => 10,
                Some(c) => return Err(ParsePrefixedU32Error::InvalidPrefix { found: c }),
                None => 10,
            };

            match radix {
                10 => Ok(self.parse()?),
                radix => Ok(u32::from_str_radix(&self[2..], radix)?),
            }
        }
    }

    #[cfg(test)]
    mod test {
        use super::ParsePrefixedU32;

        #[test]
        fn parse_prefixed_u32() {
            assert_eq!("0xdeadbeef".parse_prefixed_u32().unwrap(), 0xdeadbeef);
            assert_eq!("0xDEADBEEF".parse_prefixed_u32().unwrap(), 0xdeadbeef);
            assert_eq!("1234".parse_prefixed_u32().unwrap(), 1234);

            assert!("0abcde".parse_prefixed_u32().is_err());
        }
    }
}

mod parse_amodvalues {
    use std::str::FromStr;

    use crate::types::AmodValues;

    use super::ParseEnumError;

    impl FromStr for AmodValues {
        type Err = ParseEnumError;

        fn from_str(s: &str) -> Result<Self, Self::Err> {
            match s {
                "A16" => Ok(Self::A16),
                "A24" => Ok(Self::A24),
                "A32" => Ok(Self::A32),
                "CRCSR" => Ok(Self::CRCSR),
                "USER1" => Ok(Self::USER1),
                "USER2" => Ok(Self::USER2),
                s => Err(ParseEnumError::invalid_variant::<AmodValues>(s.to_string())),
            }
        }
    }

    #[cfg(test)]
    mod test {
        use super::AmodValues;
        use super::ParseEnumError;

        #[test]
        fn parse_amodvalues() {
            assert_eq!("A32".parse(), Ok(AmodValues::A32));
            assert_eq!(
                "A32aa".parse::<AmodValues>(),
                Err(ParseEnumError::invalid_variant::<AmodValues>(String::from(
                    "A32aa"
                )))
            );
        }
    }
}

mod parse_datatype {
    use std::str::FromStr;

    use crate::types::DataType;

    use super::ParseEnumError;

    impl FromStr for DataType {
        type Err = ParseEnumError;

        fn from_str(s: &str) -> Result<Self, Self::Err> {
            match s {
                "D32" => Ok(Self::D32),
                s => Err(ParseEnumError::invalid_variant::<Self>(s.to_string())),
            }
        }
    }
}

mod parse_rw_specifier {
    use std::str::FromStr;

    use crate::types::RwSpecifier;

    use super::ParseEnumError;

    impl FromStr for RwSpecifier {
        type Err = ParseEnumError;

        fn from_str(s: &str) -> Result<Self, Self::Err> {
            match s {
                "R" => Ok(Self::R),
                "W" => Ok(Self::W),
                "RW" => Ok(Self::RW),
                s => Err(ParseEnumError::invalid_variant::<Self>(s.to_string())),
            }
        }
    }
}

mod parse_field_fifo_interface {
    use std::str::FromStr;

    use crate::types::FieldFifoInterface;

    use super::ParseEnumError;

    impl FromStr for FieldFifoInterface {
        type Err = ParseEnumError;

        fn from_str(s: &str) -> Result<Self, Self::Err> {
            match s {
                "Vector" => Ok(Self::Vector),
                "Block" => Ok(Self::Block),
                "Both" => Ok(Self::Both),
                s => Err(ParseEnumError::invalid_variant::<Self>(s.to_string())),
            }
        }
    }
}

mod parse_custom_bool {

    use super::ParseEnumError;

    pub(crate) trait ParseCustomBool {
        fn parse_custom_bool(self) -> Result<bool, ParseEnumError>;
    }

    impl ParseCustomBool for &str {
        fn parse_custom_bool(self) -> Result<bool, ParseEnumError> {
            match self {
                "YES" | "Y" => Ok(true),
                "NO" | "N" => Ok(false),
                s => Err(ParseEnumError::invalid_variant::<bool>(s.to_string())),
            }
        }
    }

    #[cfg(test)]
    mod test {
        use super::{ParseCustomBool, ParseEnumError};

        #[test]
        fn parse_prefixed_u32() {
            assert_eq!("YES".parse_custom_bool(), Ok(true));
            assert_eq!("NO".parse_custom_bool(), Ok(false));

            assert_eq!(
                "AO".parse_custom_bool(),
                Err(ParseEnumError::invalid_variant::<bool>("AO".to_string()))
            );
        }
    }
}
